from dune.codegen.generation import (cached,
                                     class_member,
                                     get_counted_variable,
                                     global_context,
                                     include_file,
                                     preamble,
                                     )
from dune.codegen.pdelab.driver import (FEM_name_mangling,
                                        get_form_ident,
                                        get_trial_element,
                                        is_stationary,
                                        preprocess_leaf_data,
                                        )
from dune.codegen.pdelab.driver.driverblock import (name_driver_block,
                                                    type_driver_block)
from dune.codegen.pdelab.driver.gridfunctionspace import (name_trial_gfs,
                                                          name_leafview,
                                                          type_leafview,
                                                          type_range,
                                                          )
from ufl import FiniteElement, MixedElement, TensorElement, VectorElement, TensorProductElement


def interpolate_dirichlet_data(name):
    element = get_trial_element()
    func = preprocess_leaf_data(element, "interpolate_expression", applyZeroDefault=False)
    if func is not None:
        bf = name_boundary_grid_function(element, func)
        gfs = name_trial_gfs()
        interpolate_vector(bf, gfs, name)


@preamble(section="vector", kernel="driver_block")
def interpolate_vector(func, gfs, name):
    return "Dune::PDELab::interpolate(*{}, *{}, *{});".format(func,
                                                              gfs,
                                                              name,
                                                              )


@class_member(classtag="driver_block")
def typedef_composite_boundary_grid_function(name, children):
    templates = ','.join('std::decay_t<decltype(*{})>'.format(c) for c in children)
    return "using {} = Dune::PDELab::CompositeGridFunction<{}>;".format(name, templates)


def type_composite_boundary_grid_function(children, root):
    if root:
        name = "BoundaryGridFunction"
    else:
        name = "CompositeGridFunction_{}".format('_'.join(c for c in children))
    typedef_composite_boundary_grid_function(name, children)
    return name


@class_member(classtag="driver_block")
def declare_composite_boundary_grid_function(name, children, root):
    composite_gfs_type = type_composite_boundary_grid_function(children, root)
    return "std::shared_ptr<{}> {};".format(composite_gfs_type, name)


@preamble(section="vector", kernel="driver_block")
def define_composite_boundary_grid_function(name, children, root=False):
    declare_composite_boundary_grid_function(name, children, root)
    composite_gfs_type = type_composite_boundary_grid_function(children, root)
    return "{} = std::make_shared<{}>({});".format(name, composite_gfs_type, ', '.join('*{}'.format(c) for c in children))


def _is_local(func):
    # palpo TODO
    return True
    # return False
    # assert isinstance(func, tuple)
    # if len(func) == 2:
    #     return True
    # else:
    #     try:
    #         assert len(func) == 1
    #     except:
    #         from pudb import set_trace; set_trace()
    #     return False


@class_member(classtag="driver_block")
def typedef_boundary_grid_function(name, local):
    leafview_type = type_leafview()
    range_type = type_range()
    boundary_function_type = type_boundary_function(local)
    # palpo TODO: 1 in the format below!
    return "using {} = Dune::PDELab::LocalCallableToGridFunctionAdapter<{}, {}, {}, {}>;".format(name,
                                                                                                 leafview_type,
                                                                                                 range_type,
                                                                                                 1,
                                                                                                 boundary_function_type)


def type_boundary_grid_function(local, root):
    # TODO: remove local stuff
    # if local:
    #     name = "BoundaryGridFunctionLocal"
    # else:
    #     name = "BoundaryGridFunctionGlobal"
    if root:
        name = "BoundaryGridFunction"
    else:
        name = "BoundaryGridFunctionLeaf"
    typedef_boundary_grid_function(name, local)
    return name


@class_member(classtag="driver_block")
def declare_boundary_grid_function(name, local, root):
    bgf_type = type_boundary_grid_function(local, root)
    return "std::shared_ptr<{}> {};".format(bgf_type, name)


@preamble(section="vector", kernel="driver_block")
def define_boundary_grid_function(name, func, root=False):
    local = _is_local(func)
    declare_boundary_grid_function(name, local, root)
    gv = name_leafview()
    boundary_function = name_boundary_function(func, local)
    bgf_type = type_boundary_grid_function(local, root)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    if is_stationary():
        return "{} = std::make_shared<{}>({}, *{});".format(name, bgf_type, gv, boundary_function)
    else:
        # palpo TODO
        assert False
        from dune.codegen.pdelab.driver.gridoperator import name_localoperator
        lop = name_localoperator(get_form_ident())
        return "auto {} = Dune::PDELab::makeInstationaryGridFunctionFromCallable({}, {}, {});".format(name,
                                                                                                      gv,
                                                                                                      lambdaname,
                                                                                                      lop,
                                                                                                      )


@cached
def name_boundary_grid_function(element, func, root=True):
    assert isinstance(func, tuple)

    if isinstance(element, MixedElement):
        k = 0
        childs = []
        for subel in element.sub_elements():
            childs.append(name_boundary_grid_function(subel, func[k:k + subel.value_size()], root=False))
            k = k + subel.value_size()
        name = "_".join(childs)
        if len(childs) == 1:
            name = "{}_dummy".format(name)
        if root:
            name = "boundary_grid_function"
        define_composite_boundary_grid_function(name, tuple(childs), root=root)
    else:
        assert isinstance(element, (FiniteElement, TensorProductElement))
        name = get_counted_variable("boundary_grid_function")
        if root:
            name = "boundary_grid_function"
        define_boundary_grid_function(name, func, root=root)
    if root:
        print("palpo 1 element: {}".format(element))
        driver_block_get_boundarygridfunction(element, func, name=name)
    return name


def boundary_lambda(func, local):
    # palpo TODO
    assert isinstance(func, tuple)
    func = func[0]
    if func is None:
        func = 0.0

    if isinstance(func, (int, float)):
        return "[&](const auto& is, const auto& x){{ return {}; }}".format(float(func))
    else:
        from ufl.classes import Expr
        assert isinstance(func, Expr)
        from dune.codegen.pdelab.driver.visitor import ufl_to_code
        return "[&](const auto& is, const auto& xl){{ {}; }}".format(ufl_to_code(func))


@class_member(classtag="driver_block")
def typedef_boundary_function(name, local):
    range_type = type_range()
    leafview_type = type_leafview()
    if not local:
        coordinate = "typename {}::template Codim<0>::Entity::Geometry::GlobalCoordinate".format(leafview_type)
        return "using {} = std::function<{}({})>;".format(name, range_type, coordinate)
    else:
        entity = "typename {}::template Codim<0>::Entity".format(leafview_type)
        coordinate = "typename {}::template Codim<0>::Entity::Geometry::LocalCoordinate".format(leafview_type)
        return "using {} = std::function<{}({}, {})>;".format(name, range_type, entity, coordinate)


def type_boundary_function(local):
    name = "BoundaryFunction"
    if not local:
        name = name + "Global"
    else:
        name = name + "Local"
    typedef_boundary_function(name, local)
    return name


@class_member(classtag="driver_block")
def declare_boundary_function(name, local):
    bf_type = type_boundary_function(local)
    return "std::shared_ptr<{}> {};".format(bf_type, name)


@preamble(section="vector", kernel="driver_block")
def define_boundary_function(name, func, local):
    declare_boundary_function(name, local)
    bf_type = type_boundary_function(local)
    bf_lambda = boundary_lambda(func, local)
    return "{} = std::make_shared<{}> ({});".format(name, bf_type, bf_lambda)


@cached
def name_boundary_function(func, local):
    name = get_counted_variable("boundary_function")
    define_boundary_function(name, func, local)
    return name


@class_member(classtag="driver_block")
def driver_block_get_boundarygridfunction(element, func, name=None):
    if not name:
        name = name_boundary_grid_function(element, func)
    bgf_type = "BoundaryGridFunction"
    return ["std::shared_ptr<{}> getBoundaryGridFunction(){{".format(bgf_type),
            "  return {};".format(name),
            "}"]


@preamble(section="postprocessing", kernel="main")
def main_typedef_boundary_grid_function(name, func):
    driver_block_type = type_driver_block()
    local = _is_local(func)
    bgf_type = type_boundary_grid_function(local, True)
    return "using {} = {}::{};".format(name, driver_block_type, bgf_type)


def main_type_boundary_grid_function(func):
    name = "BoundaryGridFunction"
    main_typedef_boundary_grid_function(name, func)
    return name


@preamble(section="postprocessing", kernel="main")
def main_define_boundary_grid_function(name, element, func):
    driver_block_name = name_driver_block()
    print("palpo 2 element: {}".format(element))
    # driver_block_get_boundarygridfunction(element, func)
    return "auto {} = {}.getBoundaryGridFunction();".format(name, driver_block_name)


@cached
def main_name_boundary_grid_function(element, func):
    assert isinstance(func, tuple)
    name = "boundaryGridFunction"
    main_define_boundary_grid_function(name, element, func)
    return name
